"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.run = void 0;
const node_path_1 = __importDefault(require("node:path"));
const ts_morph_1 = require("ts-morph");
const ts_morph_helpers_1 = require("ts-morph-helpers");
const util_1 = require("./util");
const log_1 = require("./log");
const lineRewriter = new log_1.LineRewriter();
async function run(configuration) {
    const { cwd, isShowProgress, include, jsDocOptions } = configuration;
    const production = await (0, util_1.createProject)(cwd, configuration.entryFiles);
    const entryFiles = production.getSourceFiles();
    production.resolveSourceFileDependencies();
    const productionFiles = production.getSourceFiles();
    const project = await (0, util_1.createProject)(cwd, configuration.projectFiles);
    const projectFiles = project.getSourceFiles();
    const [usedProductionFiles, unreferencedProductionFiles] = (0, util_1.partitionSourceFiles)(projectFiles, productionFiles);
    const [, usedNonEntryFiles] = (0, util_1.partitionSourceFiles)(usedProductionFiles, entryFiles);
    const issues = {
        files: new Set(unreferencedProductionFiles.map(file => file.getFilePath())),
        exports: {},
        types: {},
        nsExports: {},
        nsTypes: {},
        duplicates: {},
    };
    const counters = {
        files: issues.files.size,
        exports: 0,
        types: 0,
        nsExports: 0,
        nsTypes: 0,
        duplicates: 0,
        processed: issues.files.size,
    };
    const updateProcessingOutput = (item) => {
        if (!isShowProgress)
            return;
        const counter = unreferencedProductionFiles.length + counters.processed;
        const total = unreferencedProductionFiles.length + usedNonEntryFiles.length;
        const percentage = Math.floor((counter / total) * 100);
        const messages = [(0, log_1.getLine)(`${percentage}%`, `of files processed (${counter} of ${total})`)];
        include.files && messages.push((0, log_1.getLine)(unreferencedProductionFiles.length, 'unused files'));
        include.exports && messages.push((0, log_1.getLine)(counters.exports, 'unused exports'));
        include.nsExports && messages.push((0, log_1.getLine)(counters.nsExports, 'unused exports in namespace'));
        include.types && messages.push((0, log_1.getLine)(counters.types, 'unused types'));
        include.nsTypes && messages.push((0, log_1.getLine)(counters.nsTypes, 'unused types in namespace'));
        include.duplicates && messages.push((0, log_1.getLine)(counters.duplicates, 'duplicate exports'));
        if (counter < total) {
            messages.push('');
            messages.push(`Processing: ${node_path_1.default.relative(cwd, item.filePath)}`);
        }
        lineRewriter.update(messages);
    };
    const addIssue = (issueType, issue) => {
        const { filePath, symbol } = issue;
        const key = node_path_1.default.relative(cwd, filePath);
        issues[issueType][key] = issues[issueType][key] ?? {};
        issues[issueType][key][symbol] = issue;
        counters[issueType]++;
        updateProcessingOutput(issue);
    };
    if (include.exports || include.types || include.nsExports || include.nsTypes || include.duplicates) {
        usedNonEntryFiles.forEach(sourceFile => {
            const filePath = sourceFile.getFilePath();
            const exportDeclarations = sourceFile.getExportedDeclarations();
            if (include.duplicates) {
                const duplicateExports = (0, ts_morph_helpers_1.findDuplicateExportedNames)(sourceFile);
                duplicateExports.forEach(symbols => {
                    const symbol = symbols.join('|');
                    addIssue('duplicates', { filePath, symbol, symbols });
                });
            }
            if (include.exports || include.types || include.nsExports || include.nsTypes) {
                const uniqueExportedSymbols = new Set([...exportDeclarations.values()].flat());
                if (uniqueExportedSymbols.size === 1)
                    return;
                exportDeclarations.forEach(declarations => {
                    declarations.forEach(declaration => {
                        const type = (0, util_1.getType)(declaration);
                        if (!include.nsExports && !include.nsTypes) {
                            if (!include.types && type)
                                return;
                            if (!include.exports && !type)
                                return;
                        }
                        if (jsDocOptions.isReadPublicTag && ts_morph_1.ts.getJSDocPublicTag(declaration.compilerNode))
                            return;
                        let identifier;
                        if (declaration.isKind(ts_morph_1.ts.SyntaxKind.Identifier)) {
                            identifier = declaration;
                        }
                        else if (declaration.isKind(ts_morph_1.ts.SyntaxKind.FunctionDeclaration) ||
                            declaration.isKind(ts_morph_1.ts.SyntaxKind.ClassDeclaration) ||
                            declaration.isKind(ts_morph_1.ts.SyntaxKind.TypeAliasDeclaration) ||
                            declaration.isKind(ts_morph_1.ts.SyntaxKind.InterfaceDeclaration) ||
                            declaration.isKind(ts_morph_1.ts.SyntaxKind.EnumDeclaration)) {
                            identifier = declaration.getFirstChildByKindOrThrow(ts_morph_1.ts.SyntaxKind.Identifier);
                        }
                        else if (declaration.isKind(ts_morph_1.ts.SyntaxKind.PropertyAccessExpression)) {
                            identifier = declaration.getLastChildByKindOrThrow(ts_morph_1.ts.SyntaxKind.Identifier);
                        }
                        else {
                            identifier = declaration.getFirstDescendantByKind(ts_morph_1.ts.SyntaxKind.Identifier);
                        }
                        if (identifier) {
                            const identifierText = identifier.getText();
                            if (include.exports && issues.exports[filePath]?.[identifierText])
                                return;
                            if (include.types && issues.types[filePath]?.[identifierText])
                                return;
                            if (include.nsExports && issues.nsExports[filePath]?.[identifierText])
                                return;
                            if (include.nsTypes && issues.nsTypes[filePath]?.[identifierText])
                                return;
                            const refs = identifier.findReferences();
                            if (refs.length === 0) {
                                addIssue('exports', { filePath, symbol: identifierText });
                            }
                            else {
                                const refFiles = new Set(refs.map(r => r.compilerObject.references.map(r => r.fileName)).flat());
                                const isReferencedOnlyBySelf = refFiles.size === 1 && [...refFiles][0] === sourceFile.getFilePath();
                                if (!isReferencedOnlyBySelf)
                                    return;
                                if ((0, ts_morph_helpers_1.findReferencingNamespaceNodes)(sourceFile).length > 0) {
                                    if (type) {
                                        addIssue('nsTypes', { filePath, symbol: identifierText, symbolType: type });
                                    }
                                    else {
                                        addIssue('nsExports', { filePath, symbol: identifierText });
                                    }
                                }
                                else if (type) {
                                    addIssue('types', { filePath, symbol: identifierText, symbolType: type });
                                }
                                else {
                                    addIssue('exports', { filePath, symbol: identifierText });
                                }
                            }
                        }
                    });
                });
            }
            counters.processed++;
        });
    }
    if (isShowProgress)
        lineRewriter.resetLines();
    return { issues, counters };
}
exports.run = run;
